"""
A command-line interface for disassembling and assembling
the Hermes Bytecode.

Usage:
    hbctool disasm <HBC_FILE> <HASM_PATH>
    hbctool asm <HASM_PATH> <HBC_FILE>
    hbctool --help
    hbctool --version

Operation:
    disasm              Disassemble Hermes Bytecode
    asm                 Assemble Hermes Bytecode

Args:
    HBC_FILE            Target HBC file
    HASM_PATH           Target HASM directory path

Options:
    --version           Show hbctool version
    --help              Show hbctool help manual

Examples:
    hbctool disasm index.android.bundle test_hasm
    hbctool asm test_hasm index.android.bundle
"""
from docopt import docopt
from hbctool import metadata, hbc, hasm

def disasm(hbcfile, hasmpath):
    print(f"[*] Disassemble '{hbcfile}' to '{hasmpath}' path")
    f = open(hbcfile, "rb")
    hbco = hbc.load(f)
    f.close()

    header = hbco.getHeader()
    sourceHash = bytes(header["sourceHash"]).hex()
    version = header["version"]
    print(f"[*] Hermes Bytecode [ Source Hash: {sourceHash}, HBC Version: {version} ]")

    hasm.dump(hbco, hasmpath)
    print(f"[*] Done")

def asm(hasmpath, hbcfile):
    print(f"[*] Assemble '{hasmpath}' to '{hbcfile}' path")
    hbco = hasm.load(hasmpath)

    header = hbco.getHeader()
    sourceHash = bytes(header["sourceHash"]).hex()
    version = header["version"]
    print(f"[*] Hermes Bytecode [ Source Hash: {sourceHash}, HBC Version: {version} ]")

    f = open(hbcfile, "wb")
    hbc.dump(hbco, f)
    f.close()
    print(f"[*] Done")

def main():
    args = docopt(__doc__, version=f"{metadata.project} {metadata.version}")
    if args['disasm']:
        disasm(args['<HBC_FILE>'], args['<HASM_PATH>'])
    elif args['asm']:
        asm(args['<HASM_PATH>'], args['<HBC_FILE>'])
    

def entry_point():
    """Zero-argument entry point for use with setuptools/distribute."""
    main()

if __name__ == "__main__":
    main()


from hbctool.util import *
from .parser import parse, export, INVALID_LENGTH
from .translator import disassemble, assemble
from struct import pack, unpack

NullTag = 0
TrueTag = 1 << 4
FalseTag = 2 << 4
NumberTag = 3 << 4
LongStringTag = 4 << 4
ShortStringTag = 5 << 4
ByteStringTag = 6 << 4
IntegerTag = 7 << 4
TagMask = 0x70

class HBC74:
    def __init__(self, f=None):
        if f:
            self.obj = parse(f)
        else:
            self.obj = None

    def export(self, f):
        export(self.getObj(), f)

    def getObj(self):
        assert self.obj, "Obj is not set."
        return self.obj

    def setObj(self, obj):
        self.obj = obj

    def getVersion(self):
        return 74    

    def getHeader(self):
        return self.getObj()["header"]

    def getFunctionCount(self):
        return self.getObj()["header"]["functionCount"]

    def getFunction(self, fid, disasm=True):
        assert fid >= 0 and fid < self.getFunctionCount(), "Invalid function ID"

        functionHeader = self.getObj()["functionHeaders"][fid]
        offset = functionHeader["offset"]
        paramCount = functionHeader["paramCount"]
        registerCount = functionHeader["frameSize"]
        symbolCount = functionHeader["environmentSize"]
        bytecodeSizeInBytes = functionHeader["bytecodeSizeInBytes"]
        functionName = functionHeader["functionName"]

        instOffset = self.getObj()["instOffset"]
        start = offset - instOffset
        end = start + bytecodeSizeInBytes
        bc = self.getObj()["inst"][start:end]
        insts = bc
        if disasm:
            insts = disassemble(bc)
        
        functionNameStr, _ = self.getString(functionName)

        return functionNameStr, paramCount, registerCount, symbolCount, insts, functionHeader
    
    def setFunction(self, fid, func, disasm=True):
        assert fid >= 0 and fid < self.getFunctionCount(), "Invalid function ID"

        functionName, paramCount, registerCount, symbolCount, insts, _ = func

        functionHeader = self.getObj()["functionHeaders"][fid]

        functionHeader["paramCount"] = paramCount
        functionHeader["frameSize"] = registerCount
        functionHeader["environmentSize"] = symbolCount

        # TODO : Make this work
        # functionHeader["functionName"] = functionName

        offset = functionHeader["offset"]
        bytecodeSizeInBytes = functionHeader["bytecodeSizeInBytes"]

        instOffset = self.getObj()["instOffset"]
        start = offset - instOffset
        
        bc = insts

        if disasm:
            bc = assemble(insts)
            
        assert len(bc) <= bytecodeSizeInBytes, "Overflowed instruction length is not supported yet."
        functionHeader["bytecodeSizeInBytes"] = len(bc)
        memcpy(self.getObj()["inst"], bc, start, len(bc))
        
    def getStringCount(self):
        return self.getObj()["header"]["stringCount"]

    def getString(self, sid):
        assert sid >= 0 and sid < self.getStringCount(), "Invalid string ID"

        stringTableEntry = self.getObj()["stringTableEntries"][sid]
        stringStorage = self.getObj()["stringStorage"]
        stringTableOverflowEntries = self.getObj()["stringTableOverflowEntries"]

        isUTF16 = stringTableEntry["isUTF16"]
        offset = stringTableEntry["offset"]
        length = stringTableEntry["length"]

        if length >= INVALID_LENGTH:
            stringTableOverflowEntry = stringTableOverflowEntries[offset]
            offset = stringTableOverflowEntry["offset"]
            length = stringTableOverflowEntry["length"]

        if isUTF16:
            length*=2

        s = bytes(stringStorage[offset:offset + length])
        return s.hex() if isUTF16 else s.decode("utf-8"), (isUTF16, offset, length)
    
    def setString(self, sid, val):
        assert sid >= 0 and sid < self.getStringCount(), "Invalid string ID"

        stringTableEntry = self.getObj()["stringTableEntries"][sid]
        stringStorage = self.getObj()["stringStorage"]
        stringTableOverflowEntries = self.getObj()["stringTableOverflowEntries"]

        isUTF16 = stringTableEntry["isUTF16"]
        offset = stringTableEntry["offset"]
        length = stringTableEntry["length"]

        if length >= INVALID_LENGTH:
            stringTableOverflowEntry = stringTableOverflowEntries[offset]
            offset = stringTableOverflowEntry["offset"]
            length = stringTableOverflowEntry["length"]
        
        if isUTF16:
            s = list(bytes.fromhex(val))
            l = len(s)//2
        else:
            l = len(val)
            s = val.encode("utf-8")
        
        assert l <= length, "Overflowed string length is not supported yet."

        memcpy(stringStorage, s, offset, len(s))
        
    def _checkBufferTag(self, buf, iid):
        keyTag = buf[iid]
        if keyTag & 0x80:
            return (((keyTag & 0x0f) << 8) | (buf[iid + 1]), keyTag & TagMask)
        else:
            return (keyTag & 0x0f, keyTag & TagMask)

    def _SLPToString(self, tag, buf, iid, ind):
        start = iid + ind
        if tag == ByteStringTag:
            type = "String"
            val = buf[start]
            ind += 1
        elif tag == ShortStringTag:
            type = "String"
            val = unpack("<H", bytes(buf[start:start+2]))[0]
            ind += 2
        elif tag == LongStringTag:
            type = "String"
            val = unpack("<L", bytes(buf[start:start+4]))[0]
            ind += 4
        elif tag == NumberTag:
            type = "Number"
            val = unpack("<d", bytes(buf[start:start+8]))[0]
            ind += 8
        elif tag == IntegerTag:
            type = "Integer"
            val = unpack("<L", bytes(buf[start:start+4]))[0]
            ind += 4
        elif tag == NullTag:
            type = "Null"
            val = None
        elif tag == TrueTag:
            type = "Boolean"
            val = True
        elif tag == FalseTag:
            type = "Boolean"
            val = False
        else:
            type = "Empty"
            val = None
        
        return type, val, ind

    def getArrayBufferSize(self):
        return self.getObj()["header"]["arrayBufferSize"]

    def getArray(self, aid):
        assert aid >= 0 and aid < self.getArrayBufferSize(), "Invalid Array ID"
        tag = self._checkBufferTag(self.getObj()["arrayBuffer"], aid)
        ind = 2 if tag[0] > 0x0f else 1
        arr = []
        t = None
        for _ in range(tag[0]):
            t, val, ind = self._SLPToString(tag[1], self.getObj()["arrayBuffer"], aid, ind)
            arr.append(val)
        
        return t, arr

    def getObjKeyBufferSize(self):
        return self.getObj()["header"]["objKeyBufferSize"]

    def getObjKey(self, kid):
        assert kid >= 0 and kid < self.getObjKeyBufferSize(), "Invalid ObjKey ID"
        tag = self._checkBufferTag(self.getObj()["objKeyBuffer"], kid)
        ind = 2 if tag[0] > 0x0f else 1
        keys = []
        t = None
        for _ in range(tag[0]):
            t, val, ind = self._SLPToString(tag[1], self.getObj()["objKeyBuffer"], kid, ind)
            keys.append(val)
        
        return t, keys

    def getObjValueBufferSize(self):
        return self.getObj()["header"]["objValueBufferSize"]

    def getObjValue(self, vid):
        assert vid >= 0 and vid < self.getObjValueBufferSize(), "Invalid ObjValue ID"
        tag = self._checkBufferTag(self.getObj()["objValueBuffer"], vid)
        ind = 2 if tag[0] > 0x0f else 1
        keys = []
        t = None
        for _ in range(tag[0]):
            t, val, ind = self._SLPToString(tag[1], self.getObj()["objValueBuffer"], vid, ind)
            keys.append(val)
        
        return t, keys



import unittest
from hbctool.util import *
from hbctool.hbc.hbc76.test import *
from hbctool.hbc.hbc74.test import *
from hbctool.hbc.hbc62.test import *
from hbctool.hbc.hbc59.test import *
from hbctool import hbc as hbcl, hasm
import pathlib
import json

basepath = pathlib.Path(__file__).parent.absolute()

class ByteIO:
    def __init__(self, v=b""):
        self.buf = v

    def write(self, b):
        self.buf += b

    def read(self, n=-1):
        if n==-1:
            o = self.buf
            self.buf = b""
            return o

        o = self.buf[:n]
        self.buf = self.buf[n:]
        return o

class TestFileUtilization(unittest.TestCase):
    def test_bit_writer(self):
        io = ByteIO()
        fw = BitWriter(io)

        offset = 182856
        paramCount = 1
        bytecodeSizeInBytes = 12342
        functionName = 3086

        write(fw, offset, ["bit", 25, 1])
        write(fw, paramCount, ["bit", 7, 1])
        write(fw, bytecodeSizeInBytes, ["bit", 15, 1])
        write(fw, functionName, ["bit", 17, 1])

        bs = io.read()
        self.assertEqual("48ca020236300706", bs.hex())

        isUTF16 = 0
        offset = 465
        length = 3

        write(fw, isUTF16, ["bit", 1, 1])
        write(fw, offset, ["bit", 23, 1])
        write(fw, length, ["bit", 8, 1])

        bs = io.read()
        self.assertEqual("a2030003", bs.hex())

    def test_bit_reader(self):
        io = ByteIO()
        fr = BitReader(io)

        io.write(bytes.fromhex("48ca020236300706"))

        offset = read(fr, ["bit", 25, 1])
        paramCount = read(fr, ["bit", 7, 1])
        bytecodeSizeInBytes = read(fr, ["bit", 15, 1])
        functionName = read(fr, ["bit", 17, 1])

        self.assertEqual(offset, 182856)
        self.assertEqual(paramCount, 1)
        self.assertEqual(bytecodeSizeInBytes, 12342)
        self.assertEqual(functionName, 3086)

        io.write(bytes.fromhex("a2030003"))

        isUTF16 = read(fr, ["bit", 1, 1])
        offset = read(fr, ["bit", 23, 1])
        length = read(fr, ["bit", 8, 1])

        self.assertEqual(isUTF16, 0)
        self.assertEqual(offset, 465)
        self.assertEqual(length, 3)

    def test_conversion(self):

        io = ByteIO()
        fr = BitReader(io)
        fw = BitWriter(io)

        isUTF16 = 1
        offset = 465
        length = 3

        write(fw, isUTF16, ["bit", 1, 1])
        write(fw, offset, ["bit", 23, 1])
        write(fw, length, ["bit", 8, 1])

        isUTF16 = read(fr, ["bit", 1, 1])
        offset = read(fr, ["bit", 23, 1])
        length = read(fr, ["bit", 8, 1])

        self.assertEqual(isUTF16, 1)
        self.assertEqual(offset, 465)
        self.assertEqual(length, 3)

def main():
    unittest.main()

if __name__ == "__main__":
    main()




from hbctool.util import *
from hbctool.hbc.hbc76 import HBC76
from hbctool.hbc.hbc74 import HBC74
from hbctool.hbc.hbc62 import HBC62
from hbctool.hbc.hbc59 import HBC59
import json

MAGIC = 2240826417119764422
INIT_HEADER = {
    "magic": ["uint", 64, 1],
    "version": ["uint", 32, 1]
}
BYTECODE_ALIGNMENT = 4

HBC = {
    76: HBC76,
    74: HBC74,
    62: HBC62,
    59: HBC59
}

def load(f):
    f = BitReader(f)
    magic = read(f, INIT_HEADER["magic"])
    version = read(f, INIT_HEADER["version"])
    f.seek(0)
    assert magic == MAGIC, f"The magic ({hex(magic)}) is invalid. (must be {hex(MAGIC)})"
    assert version in HBC, f"The HBC version ({version}) is not supported."

    return HBC[version](f)

def loado(obj):
    magic = obj["header"]["magic"]
    version = obj["header"]["version"]

    assert magic == MAGIC, f"The magic ({hex(magic)}) is invalid. (must be {hex(MAGIC)})"
    assert version in HBC, f"The HBC version ({version}) is not supported."

    hbc = HBC[version]()
    hbc.setObj(obj)
    return hbc

def dump(hbc, f):
    f = BitWriter(f)
    hbc.export(f)

def dumpo(hbc):
    return hbc.getObj()


import pathlib
import json
from hbctool.util import *

basepath = pathlib.Path(__file__).parent.absolute()

operand_type = {
    "Reg8": (1, to_uint8, from_uint8),
    "Reg32": (4, to_uint32, from_uint32),
    "UInt8": (1, to_uint8, from_uint8),
    "UInt16": (2, to_uint16, from_uint16),
    "UInt32": (4, to_uint32, from_uint32),
    "Addr8": (1, to_int8, from_int8),
    "Addr32": (4, to_int32, from_int32),
    "Reg32": (4, to_uint32, from_uint32),
    "Imm32": (4, to_uint32, from_uint32),
    "Double": (8, to_double, from_double)
}

f = open(f"{basepath}/data/opcode.json", "r")
opcode_operand = json.load(f)
opcode_mapper = list(opcode_operand.keys())
opcode_mapper_inv = {}
for i, v in enumerate(opcode_mapper):
    opcode_mapper_inv[v] = i

f.close()

def disassemble(bc):
    i = 0
    insts = []
    while i < len(bc):
        opcode = opcode_mapper[bc[i]]
        i+=1
        inst = (opcode, [])
        operand_ts = opcode_operand[opcode]
        for oper_t in operand_ts:
            is_str = oper_t.endswith(":S")
            if is_str:
                oper_t = oper_t[:-2]
                
            size, conv_to, _ = operand_type[oper_t]
            val = conv_to(bc[i:i+size])
            inst[1].append((oper_t, is_str, val))
            i+=size
        
        insts.append(inst)
        
    return insts

def assemble(insts):
    bc = []
    for opcode, operands in insts:
        op = opcode_mapper_inv[opcode]
        bc.append(op)
        assert len(opcode_operand[opcode]) == len(operands), f"Malicious instruction: {op}, {operands}"
        for oper_t, _, val in operands:
            assert oper_t in operand_type, f"Malicious operand type: {oper_t}"
            _, _, conv_from = operand_type[oper_t]
            bc += conv_from(val)
    
    return bc


package = 'hbctool'
project = 'hbctool'
project_no_spaces = project.replace(' ', '')
version = '0.1.5'
description = ('A command-line interface for disassembling and assembling'
               'the Hermes Bytecode')
authors = ['bongtrop', 'ErbaZZ']
authors_string = ', '.join(authors)
emails = ['bongtrop@gmail.com', 'weerawat.paw@gmail.com']
license = 'MIT'
url = 'https://github.com/bongtrop/hbctool'


from hbctool import hbc as hbcl, hasm
from .translator import assemble, disassemble
import unittest
import re
import pathlib
import json

basepath = pathlib.Path(__file__).parent.absolute()

class TestHBC59(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestHBC59, self).__init__(*args, **kwargs)
        self.hbc = hbcl.load(open(f"{basepath}/example/index.android.bundle", "rb"))
        # self.objdump = open(f"{basepath}/example/objdump.out", "r").read()
        self.pretty = open(f"{basepath}/example/pretty.out", "r").read()
        self.raw = open(f"{basepath}/example/raw.out", "r").read()

    def test_get_function(self):
        # hbcdump version 59 cannot fetch
        # target_offsets = re.findall(r"([0-9a-f]+) \<_[0-9]+\>", self.objdump)
        target_args = re.findall(r"Function<(.*?)>([0-9]+)\(([0-9]+) params, ([0-9]+) registers,\s?([0-9]+) symbols\):", self.pretty)

        functionCount = self.hbc.getFunctionCount()

        # self.assertEqual(functionCount, len(target_offsets))
        self.assertEqual(functionCount, len(target_args))

        for i in range(functionCount):
            # target_offset = target_offsets[i]
            target_functionName, _, target_paramCount, target_registerCount, target_symbolCount = target_args[i]

            try:
                functionName, paramCount, registerCount, symbolCount, _, funcHeader = self.hbc.getFunction(i)
            except AssertionError:
                self.fail()

            self.assertEqual(functionName, target_functionName)
            self.assertEqual(paramCount, int(target_paramCount))
            self.assertEqual(registerCount, int(target_registerCount))
            self.assertEqual(symbolCount, int(target_symbolCount))
            # self.assertEqual(funcHeader["offset"], int(target_offset, 16))
    
    def test_get_string(self):
        target_strings = re.findall(r"[isp][0-9]+\[([UTFASCI16-]+), ([0-9]+)..([0-9-]+)\].*?:\s?(.*)", self.pretty)
        stringCount = self.hbc.getStringCount()

        self.assertEqual(stringCount, len(target_strings))

        for i in range(stringCount):
            val, header = self.hbc.getString(i)
            isUTF16, offset, length = header

            t, target_start, target_end, target_val = target_strings[i]

            target_isUTF16 = t == "UTF-16"
            target_offset = int(target_start)
            target_length = int(target_end) - target_offset + 1

            self.assertEqual(isUTF16, target_isUTF16)
            self.assertEqual(offset, target_offset)
            self.assertEqual(length, target_length)

            # TODO : Implement this please
            # self.assertEqual(val, target_val)

    def test_translator(self):
        functionCount = self.hbc.getFunctionCount()

        for i in range(functionCount):
            _, _, _, _, bc, _ = self.hbc.getFunction(i, disasm=False)

            self.assertEqual(assemble(disassemble(bc)), bc)
class TestParser59(unittest.TestCase):
    def test_hbc(self):
        f = open(f"{basepath}/example/index.android.bundle", "rb")
        hbc = hbcl.load(f)
        f.close()
        f = open("/tmp/hbctool_test.android.bundle", "wb")
        hbcl.dump(hbc, f)
        f.close()

        f = open("hbc/hbc59/example/index.android.bundle", "rb")
        a = f.read()
        f.close()
        f = open("/tmp/hbctool_test.android.bundle", "rb")
        b = f.read()
        f.close()

        self.assertEqual(a, b)

    def test_hasm(self):
        f = open(f"{basepath}/example/index.android.bundle", "rb")
        a = hbcl.load(f)
        f.close()
        hasm.dump(a, "/tmp/hbctool_test", force=True)
        b = hasm.load("/tmp/hbctool_test")

        self.assertEqual(json.dumps(a.getObj()), json.dumps(b.getObj()))


from .util import *
import hbctool.hbc as hbcl
import json
import os
import shutil
import re

def write_func(f, func, i, hbc):
    functionName, paramCount, registerCount, symbolCount, insts, _ = func
    f.write(f"Function<{functionName}>{i}({paramCount} params, {registerCount} registers, {symbolCount} symbols):\n")
    for opcode, operands in insts:
        f.write(f"\t{opcode.ljust(20,' ')}\t")
        o = []
        ss = []
        for ii, v in enumerate(operands):
            t, is_str, val = v
            o.append(f"{t}:{val}")

            if is_str:
                s, _ = hbc.getString(val)
                ss.append((ii, val, s))
                
        
        f.write(f"{', '.join(o)}\n")
        if len(ss) > 0:
            for ii, val, s in ss:
                f.write(f"\t; Oper[{ii}]: String({val}) {repr(s)}\n")

            f.write("\n")

    f.write("EndFunction\n\n")

def dump(hbc, path, force=False):
    
    if os.path.exists(path) and not force:
        c = input(f"'{path}' exists. Do you want to remove it ? (y/n): ").lower().strip()
        if c[:1] == "y":
            shutil.rmtree(path)
        else:
            exit(1337)
    
    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path)
    # Write all obj to metadata.json
    f = open(f"{path}/metadata.json", "w")
    json.dump(hbc.getObj(), f)
    f.close()
    
    stringCount = hbc.getStringCount()
    functionCount = hbc.getFunctionCount()

    ss = []
    for i in range(stringCount):
        val, header = hbc.getString(i)
        ss.append({
            "id": i,
            "isUTF16": header[0] == 1,
            "value": val
        })
    
    f = open(f"{path}/string.json", "w")
    json.dump(ss, f, indent=4)
    f.close()

    f = open(f"{path}/instruction.hasm", "w")
    for i in range(functionCount):
        write_func(f, hbc.getFunction(i), i, hbc)
    f.close()

def read_all_func(hasm, hbc):
    func_asms = [func_asm + "EndFunction" for func_asm in hasm.split("EndFunction\n\n")[:-1]]
    functionCount = hbc.getFunctionCount()

    rs = [''] * functionCount

    for func_asm in func_asms:
        m = re.search(r"Function<.*?>([0-9]+)\([0-9]+ params, [0-9]+ registers,\s?[0-9]+ symbols\):", func_asm)
        assert m, f"Malicious function header: {func_asm}"

        fid = int(m.group(1))

        assert fid >= 0 and fid < functionCount, f"Malicious function ID: {fid} (must lower than {functionCount})"

        rs[fid] = func_asm
    
    return rs


def read_func(func_asms, i):
    func_asm = func_asms[i]

    m = re.search(r"Function<.*?>([0-9]+)\(([0-9]+) params, ([0-9]+) registers,\s?([0-9]+) symbols\):\n(.+?)\nEndFunction", func_asm, re.DOTALL)
    assert m, f"Malicious function header: {func_asm}"

    functionName = m.group(1)
    paramCount = int(m.group(2))
    registerCount = int(m.group(3))
    symbolCount = int(m.group(4))
    insts_asm = m.group(5)

    inst_lines = insts_asm.split("\n")

    insts = []

    for inst_line in inst_lines:
        inst_line = inst_line.strip()

        if len(inst_line) == 0 or inst_line.startswith(";"):
            continue

        inst_words = inst_line.split()

        opcode = inst_words[0]

        operands = []
        for oper in inst_words[1:]:
            oper_t, val = oper.replace(",", "").split(":")
            
            if oper_t == 'Double':
                val = float(val)
            else:
                val = int(val)
            
            operands.append((oper_t, False, val))
        
        insts.append((opcode, operands))
    
    return functionName, paramCount, registerCount, symbolCount, insts, None


def load(path):
    assert os.path.exists(path), f"{path} does not exists."
    assert os.path.exists(f"{path}/metadata.json"), f"metadata.json not found."
    assert os.path.exists(f"{path}/string.json"), f"string.json not found."
    assert os.path.exists(f"{path}/instruction.hasm"), f"instruction.hasm not found."

    f = open(f"{path}/metadata.json", "r")
    hbc = hbcl.loado(json.load(f))
    f.close()

    f = open(f"{path}/instruction.hasm", "r")
    hasm_content = f.read()
    f.close()

    f = open(f"{path}/string.json", "r")
    strings = json.load(f)
    f.close()

    for string in strings:
        hbc.setString(string["id"], string["value"])
  
    func_asms = read_all_func(hasm_content, hbc)
    for i in range(len(func_asms)):
        func = read_func(func_asms, i)
        hbc.setFunction(i, func)

        
    return hbc





from struct import pack, unpack

# File Object

class BitWriter(object):
    def __init__(self, f):
        self.accumulator = 0
        self.bcount = 0
        self.out = f
        self.write = 0
        self.remained = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flush()

    def __del__(self):
        try:
            self.flush()
        except ValueError:   # I/O operation on closed file.
            pass

    def _writebit(self, bit, remaining=-1):
        if remaining > -1:
            self.accumulator |= bit << (remaining - 1)
        else:
            self.accumulator |= bit << (7 - self.bcount + self.remained)
        
        self.bcount += 1

        if self.bcount == 8:
            self.flush()

    def _clearbits(self, remaining):
        self.remained = remaining
    
    def _writebyte(self, b):
        assert not self.bcount, "bcount is not zero."
        self.out.write(bytes([b]))
        self.write += 1

    def writebits(self, v, n, remained=False):
        i = n
        while i > 0:
            self._writebit((v & (1 << i-1)) >> (i-1), remaining=(i if remained else -1))
            i -= 1
        
        if remained:
            self._clearbits(n)

    def writebytes(self, v, n):
        while n > 0:
            self._writebyte(v & 0xff)
            v = v >> 8
            n -= 1
        
        return v

    def flush(self):
        self.out.write(bytes([self.accumulator]))
        self.accumulator = 0
        self.bcount = 0
        self.remained = 0
        self.write += 1

    def seek(self, i):
        self.out.seek(i)
        self.write = i

    def tell(self):
        return self.write

    def pad(self, alignment):
        assert alignment > 0 and alignment <= 8 and ((alignment & (alignment - 1)) == 0), "Support alignment as many as 8 bytes."
        l = self.tell()
        if l % alignment == 0:
            return

        b = alignment - (l % alignment)
        self.writeall([0] * (b))
    
    def writeall(self, bs):
        self.out.write(bytes(bs))
        self.write += len(bs)

class BitReader(object):
    def __init__(self, f):
        self.input = f
        self.accumulator = 0
        self.bcount = 0
        self.read = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _readbit(self, remaining=-1):
        if not self.bcount:
            a = self.input.read(1)
            self.read += 1
            if a:
                self.accumulator = ord(a)
            self.bcount = 8

        if remaining > -1:
            assert remaining <= self.bcount, f"WTF ({remaining}, {self.bcount})"
            return (self.accumulator & (1 << remaining-1)) >> remaining-1

        rv = (self.accumulator & (1 << self.bcount-1)) >> self.bcount-1
        self.bcount -= 1
        return rv

    def _clearbits(self, remaining):
        self.bcount -= remaining
        self.accumulator = self.accumulator >> remaining

    def _readbyte(self):
        assert not self.bcount, "bcount is not zero."
        a = self.input.read(1)
        self.read += 1
        return ord(a)

    def readbits(self, n, remained=False):
        v = 0
        i = n
        while i > 0:
            v = (v << 1) | self._readbit(remaining=(i if remained else -1))
            i -= 1
        
        if remained:
            self._clearbits(n)
        
        return v
    
    def readbytes(self, n=1):
        v = 0
        while n > 0:
            v = (v << 8) | self._readbyte()
            n -= 1
        
        return v

    def seek(self, i):
        self.input.seek(i)
        self.read = i
    
    def tell(self):
        return self.read
    
    def pad(self, alignment):
        assert alignment > 0 and alignment <= 8 and ((alignment & (alignment - 1)) == 0), "Support alignment as many as 8 bytes."
        l = self.tell()
        if l % alignment == 0:
            return

        b = alignment - (l % alignment)
        self.seek(l + b)
    
    def readall(self):
        a = self.input.read()
        self.read += len(a)
        return list(a)

# File utilization function
# Read
def readuint(f, bits=64, signed=False):
    assert bits % 8 == 0, "Not support"
    if bits == 8:
        return f.readbytes(1)

    x = 0
    s = 0
    for _ in range(bits//8):
        b = f.readbytes(1)
        x |= (b & 0xFF) << s
        s += 8
    
    if signed and (x & (1<<(bits-1))):
        x = - ((1<<(bits)) - x)
        
    if x.bit_length() > bits:
        print(f"--> Int {x} longer than {bits} bits")
    return x

def readint(f, bits=64):
    return readuint(f, bits, signed=True)

def readbits(f, bits=8):
    x = 0
    s = 0

    if f.bcount % 8 != 0 and bits >= f.bcount:
        l = f.bcount
        b = f.readbits(l)
        x |= (b & 0xFF) << s
        s += l
        bits -= l
        
    for _ in range(bits//8):
        b = f.readbits(8)
        x |= (b & 0xFF) << s
        s += 8
    
    r = bits % 8
    if r != 0:
        b = f.readbits(r, remained=True)
        x |= (b & ((1 << r) - 1)) << s
        s+=r

    return x

def read(f, format):
    type = format[0]
    bits = format[1]
    n = format[2]
    r = []
    for i in range(n):
        if type == "uint":
            r.append(readuint(f, bits=bits))
        elif type == "int":
            r.append(readint(f, bits=bits))
        elif type == "bit":
            r.append(readbits(f, bits=bits))
        else:
            raise Exception(f"Data type {type} is not supported.")
    
    if len(r) == 1:
        return r[0]
    else:
        return r

# Write
def writeuint(f, v, bits=64, signed=False):
    assert bits % 8 == 0, "Not support"

    if signed:
        v += (1 << bits)

    if bits == 8:
        f.writebytes(v, 1)
        return

    s = 0
    for _ in range(bits//8):
        f.writebytes(v & 0xff, 1)
        v = v >> 8
        s += 8

def writeint(f, v, bits=64):
    return writeuint(f, v, bits, signed=True)

def writebits(f, v, bits=8):
    s = 0
    if f.bcount % 8 != 0 and bits >= 8 - f.bcount:
        l = 8 - f.bcount
        f.writebits(v & ((1 << l) - 1), l)
        v = v >> l
        s += l
        bits -= l
        
    for _ in range(bits//8):
        f.writebits(v & 0xff, 8)
        v = v >> 8
        s += 8
    
    r = bits % 8
    if r != 0:
        f.writebits(v & ((1 << bits) - 1), r, remained=True)
        v = v >> r
        s+=r

def write(f, v, format):
    t = format[0]
    bits = format[1]
    n = format[2]

    if not isinstance(v, list):
        v = [v]

    for i in range(n):
        if t == "uint":
            writeuint(f, v[i], bits=bits)
        elif t == "int":
            writeint(f, v[i], bits=bits)
        elif t == "bit":
            writebits(f, v[i], bits=bits)
        else:
            raise Exception(f"Data type {t} is not supported.")
    
# Unpacking
def to_uint8(buf):
    return buf[0]

def to_uint16(buf):
    return unpack("<H", bytes(buf[:2]))[0]

def to_uint32(buf):
    return unpack("<L", bytes(buf[:4]))[0]

def to_int8(buf):
    return unpack("<b", bytes([buf[0]]))[0]

def to_int32(buf):
    return unpack("<i", bytes(buf[:4]))[0]

def to_double(buf):
    return unpack("<d", bytes(buf[:8]))[0]

# Packing

def from_uint8(val):
    return [val]

def from_uint16(val):
    return list(pack("<H", val))

def from_uint32(val):
    return list(pack("<L", val))

def from_int8(val):
    return list(pack("<b", val))

def from_int32(val):
    return list(pack("<i", val))

def from_double(val):
    return list(pack("<d", val))

# Buf Function

def memcpy(dest, src, start, length):
    for i in range(length):
        dest[start + i] = src[i]




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hbctool',
 'hbctool.hbc',
 'hbctool.hbc.hbc59',
 'hbctool.hbc.hbc59.tool',
 'hbctool.hbc.hbc62',
 'hbctool.hbc.hbc62.tool',
 'hbctool.hbc.hbc74',
 'hbctool.hbc.hbc74.tool',
 'hbctool.hbc.hbc76',
 'hbctool.hbc.hbc76.tool']

package_data = \
{'': ['*'],
 'hbctool.hbc.hbc59': ['data/*', 'example/*', 'raw/*'],
 'hbctool.hbc.hbc62': ['data/*', 'example/*', 'raw/*'],
 'hbctool.hbc.hbc74': ['data/*', 'example/*', 'raw/*'],
 'hbctool.hbc.hbc76': ['data/*', 'example/*', 'raw/*']}

install_requires = \
['docopt>=0.6.2,<0.7.0']

entry_points = \
{'console_scripts': ['hbctool = hbctool:main']}

setup_kwargs = {
    'name': 'hbctool',
    'version': '0.1.5',
    'description': 'A command-line interface for disassembling and assembling the Hermes Bytecode.',
    'long_description': "# hbctool \n\n[![Python 3.x](https://img.shields.io/badge/python-3.x-yellow.svg)](https://python.org) [![PyPI version](https://badge.fury.io/py/hbctool.svg)](https://badge.fury.io/py/hbctool) [![Software License](https://img.shields.io/badge/license-MIT-brightgreen.svg)](/LICENSE)\n\nA command-line interface for disassembling and assembling the Hermes Bytecode.\n\nSince the React Native team created their own JavaScript engine (named Hermes) for running the React Native application, the JavaScript source code is often compiled to the Hermes bytecode. In the penetration test project, I found that some React Native applications have already been migrated to the Hermes engine. It is really head for me to analyze or patch those applications. Therefore, I created hbctool for helping any pentester to test the Hermes bytecode.\n\n> [Hermes](https://hermesengine.dev/) is an open-source JavaScript engine optimized for running React Native apps on Android. For many apps, enabling Hermes will result in improved start-up time, decreased memory usage, and smaller app size. At this time Hermes is an opt-in React Native feature, and this guide explains how to enable it.\n\nSpecial thanks to [ErbaZZ](https://github.com/ErbaZZ) and [Jusmistic](https://github.com/Jusmistic) for helping me research and develop this tool.\n\nFor more information, please visit:\n\n[https://suam.wtf/posts/react-native-application-static-analysis-en/](https://suam.wtf/posts/react-native-application-static-analysis-en/)\n\n## Screenshot\n\n![hbctool Example](/image/hbctool_example.gif)\n\nThis video with MP4 format can be found at [/image/hbctool_example.mp4](/image/hbctool_example.mp4).\n\n## Installation\n\nTo install hbctool, simply use pip:\n\n```\npip install hbctool\n```\n\n## Usage\n\nPlease run `hbctool --help` to show the usage.\n\n```\nhbctool --help   \nA command-line interface for disassembling and assembling\nthe Hermes Bytecode.\n\nUsage:\n    hbctool disasm <HBC_FILE> <HASM_PATH>\n    hbctool asm <HASM_PATH> <HBC_FILE>\n    hbctool --help\n    hbctool --version\n\nOperation:\n    disasm              Disassemble Hermes Bytecode\n    asm                 Assemble Hermes Bytecode\n\nArgs:\n    HBC_FILE            Target HBC file\n    HASM_PATH           Target HASM directory path\n\nOptions:\n    --version           Show hbctool version\n    --help              Show hbctool help manual\n\nExamples:\n    hbctool disasm index.android.bundle test_hasm\n    hbctool asm test_hasm index.android.bundle\n```\n\n> For Android, the HBC file normally locates at `assets` directory with `index.android.bundle` filename.\n\n## Support\n\nhbctool currently supports the following Hermes Bytecode version:\n\n- [Hermes Bytecode version 59](/hbctool/hbc/hbc59/)\n- [Hermes Bytecode version 62](/hbctool/hbc/hbc62/)\n- [Hermes Bytecode version 74](/hbctool/hbc/hbc74/)\n- [Hermes Bytecode version 76](/hbctool/hbc/hbc76/)\n\n## Contribution\n\nFeel free to create an issue or submit the merge request. Anyway you want to contribute this project. I'm very happy about it.\n\nHowever, please run the unit test before submiting the pull request.\n\n```\ncd hbctool\npython test.py\n```\n\nI use poetry to build this tool. To build it yourself, simply execute:\n\n```\npoetry install\n```\n\n## Next Step\n\n- Add the other Hermes bytecode versions\n- Create a class abstraction\n- Support overflow patching\n- Do all TODO, NOTE, FIXME in source code\n",
    'author': 'bongtrop',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bongtrop/hbctool',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import pathlib
import re
import json

basepath = pathlib.Path(__file__).parent.absolute()

# Open file
bytecodeListFile = open(f"{basepath}/../raw/BytecodeList.def", "r")
lines = bytecodeListFile.readlines()
bytecodeListFile.close()

# Init constants
jmp_operand = {
    "1": ["Addr8"],
    "1Long": ["Addr32"],
    "2": ["Addr8", "Reg8"],
    "2Long": ["Addr32", "Reg8"],
    "3": ["Addr8", "Reg8", "Reg8"],
    "3Long": ["Addr32", "Reg8", "Reg8"]
}

# Init variables
json_op = {}
opcode = 0
line_num  = 0

def addOp(name, operands):
    global opcode
    global json_op

    print(hex(opcode)[2:], name)
    json_op[name] = operands
    opcode = opcode + 1
    pass

# Read each line
for line in lines:
    line_num = line_num + 1

    # Example: DEFINE_OPCODE_4(NewArrayWithBuffer, Reg8, UInt16, UInt16, UInt16)
    # name = "NewArrayWithBuffer"
    # operands = ["Reg8", "UInt16", "UInt16", "UInt16"]
    if line.startswith("DEFINE_OPCODE_"):
        match = re.search(r'\((\w+)((, \w+)*)\)', line)
        name = match.group(1)
        operands = match.group(2).split(', ')[1:]
        addOp(name, operands)

    elif line.startswith("OPERAND_STRING_ID"):
        match = re.search(r'\((\w+), (\w+)\)', line)
        name = match.group(1)
        operandID = int(match.group(2)) - 1 # Starts from 0
        
        # Handle name not found / arg not found
        assert json_op[name], f"Opcode not found ({name})"
        assert json_op[name][operandID], f"Operand  not found ({operandID})"
        
        # Add ":S" to argument arg of json_op[name]
        json_op[name][operandID] += ":S"

    elif line.startswith("DEFINE_JUMP_"):
        match = re.search(r'(\d)\((\w+)\)', line)
        num_op = match.group(1)
        name = match.group(2)
        
        addOp(name, jmp_operand[f"{num_op}"])
        addOp(f"{name}Long", jmp_operand[f"{num_op}Long"])

    # NOTE : Skip all assert
    elif line.startswith("ASSERT_"):
        pass

    # NOTE : Skip this
    elif line.startswith("DEFINE_RET_TARGET"):
        pass

    # NOTE : Also skip this
    elif line.startswith("DEFINE_OPERAND_TYPE"):
        pass

    # NOTE : Skip all comments
    elif line.startswith("#") or line.startswith("//") or line.startswith("/*") or line.startswith(" *") or line.startswith("#") or line.startswith("  ") or line.startswith("\n"):
        pass
    
    else:
        # Unhandled cases
        print(line_num, line)

# Write json to file
f = open(f"{basepath}/../data/opcode.json", "w")
json.dump(json_op, f, indent=4)
f.close()



from hbctool.util import *
import json
import pathlib
import copy

basepath = pathlib.Path(__file__).parent.absolute()

MAGIC = 2240826417119764422
BYTECODE_ALIGNMENT = 4

INVALID_OFFSET = (1 << 23)
INVALID_LENGTH = (1 << 8) - 1

structure = json.load(open(f"{basepath}/data/structure.json", "r"))

headerS = structure["header"]
smallFunctionHeaderS = structure["SmallFuncHeader"]
functionHeaderS = structure["FuncHeader"]
stringTableEntryS = structure["SmallStringTableEntry"]
overflowStringTableEntryS = structure["OverflowStringTableEntry"]
stringStorageS = structure["StringStorage"]
arrayBufferS = structure["ArrayBuffer"]
objKeyBufferS = structure["ObjKeyBuffer"]
objValueBufferS = structure["ObjValueBuffer"]
regExpTableEntryS = structure["RegExpTableEntry"]
regExpStorageS = structure["RegExpStorage"]
cjsModuleTableS = structure["CJSModuleTable"]

def align(f):
    f.pad(BYTECODE_ALIGNMENT)

def parse(f):
    obj = {}

    # Segment 1: Header
    header = {}
    for key in headerS:
        header[key] = read(f, headerS[key])
    
    obj["header"] = header
    align(f)
    
    # Segment 2: Function Header
    functionHeaders = []
    for i in range(header["functionCount"]):
        functionHeader = {}
        for key in smallFunctionHeaderS:
            functionHeader[key] = read(f, smallFunctionHeaderS[key])
        
        if (functionHeader["flags"] >> 5) & 1:
            functionHeader["small"] = copy.deepcopy(functionHeader)
            saved_pos = f.tell()
            large_offset = (functionHeader["infoOffset"] << 16 )  | functionHeader["offset"]
            f.seek(large_offset)
            for key in functionHeaderS:
                functionHeader[key] = read(f, functionHeaderS[key])

            f.seek(saved_pos)
            
        functionHeaders.append(functionHeader)

    obj["functionHeaders"] = functionHeaders
    align(f)

    # Segment 3: StringKind
    # FIXME : Do nothing just skip
    stringKinds = []
    for _ in range(header["stringKindCount"]):
        stringKinds.append(readuint(f, bits=32))

    obj["stringKinds"] = stringKinds
    align(f)

    # Segment 3: IdentifierHash
    # FIXME : Do nothing just skip
    identifierTranslations = []
    for _ in range(header["identifierCount"]):
        identifierTranslations.append(readuint(f, bits=32))
    
    obj["identifierTranslations"] = identifierTranslations
    align(f)

    # Segment 4: StringTable
    stringTableEntries = []
    for _ in range(header["stringCount"]):
        stringTableEntry = {}
        for key in stringTableEntryS:
            stringTableEntry[key] = read(f, stringTableEntryS[key])
        
        stringTableEntries.append(stringTableEntry)

    obj["stringTableEntries"] = stringTableEntries
    align(f)

    # Segment 5: StringTableOverflow
    stringTableOverflowEntries = []
    for _ in range(header["overflowStringCount"]):
        stringTableOverflowEntry = {}
        for key in overflowStringTableEntryS:
            stringTableOverflowEntry[key] = read(f, overflowStringTableEntryS[key])
        
        stringTableOverflowEntries.append(stringTableOverflowEntry)
    
    obj["stringTableOverflowEntries"] = stringTableOverflowEntries
    align(f)

    # Segment 6: StringStorage
    stringStorageS[2] = header["stringStorageSize"]
    stringStorage = read(f, stringStorageS)

    obj["stringStorage"] = stringStorage
    align(f)

    # Segment 7: ArrayBuffer
    arrayBufferS[2] = header["arrayBufferSize"]
    arrayBuffer = read(f, arrayBufferS)

    obj["arrayBuffer"] = arrayBuffer
    align(f)

    # Segment 9: ObjKeyBuffer
    objKeyBufferS[2] = header["objKeyBufferSize"]
    objKeyBuffer = read(f, objKeyBufferS)

    obj["objKeyBuffer"] = objKeyBuffer
    align(f)

    # Segment 10: ObjValueBuffer
    objValueBufferS[2] = header["objValueBufferSize"]
    objValueBuffer = read(f, objValueBufferS)

    obj["objValueBuffer"] = objValueBuffer
    align(f)

    # Segment 11: RegExpTable
    regExpTable = []
    for _ in range(header["regExpCount"]):
        regExpEntry = {}
        for key in regExpTableEntryS:
            regExpEntry[key] = read(f, regExpTableEntryS[key])
        
        regExpTable.append(regExpEntry)

    obj["regExpTable"] = regExpTable
    align(f)    
    
    # Segment 12: RegExpStorage
    regExpStorageS[2] = header["regExpStorageSize"]
    regExpStorage = read(f, regExpStorageS)

    obj["regExpStorage"] = regExpStorage
    align(f)

    # Segment 13: CJSModuleTable
    cjsModuleTable = []
    for _ in range(header["cjsModuleCount"]):
        cjsModuleEntry = {}
        for key in cjsModuleTableS:
            cjsModuleEntry[key] = read(f, cjsModuleTableS[key])
        
        cjsModuleTable.append(cjsModuleEntry)

    obj["cjsModuleTable"] = cjsModuleTable
    align(f)

    obj["instOffset"] = f.tell()
    obj["inst"] = f.readall()

    return obj

def export(obj, f):
    # Segment 1: Header
    header = obj["header"]
    for key in headerS:
        write(f, header[key], headerS[key])
    
    align(f)
    
    overflowedFunctionHeaders = []
    # Segment 2: Function Header
    functionHeaders = obj["functionHeaders"]
    for i in range(header["functionCount"]):
        functionHeader = functionHeaders[i]
        if "small" in functionHeader:
            for key in smallFunctionHeaderS:
                write(f, functionHeader["small"][key], smallFunctionHeaderS[key])
            
            overflowedFunctionHeaders.append(functionHeader)
        
        else:
            for key in smallFunctionHeaderS:
                write(f, functionHeader[key], smallFunctionHeaderS[key])

    align(f)

    # Segment 3: StringKind
    # FIXME : Do nothing just skip
    stringKinds = obj["stringKinds"]
    for i in range(header["stringKindCount"]):
        writeuint(f, stringKinds[i], bits=32)

    align(f)

    # Segment 3: IdentifierHash
    # FIXME : Do nothing just skip
    identifierTranslations = obj["identifierTranslations"]
    for i in range(header["identifierCount"]):
        writeuint(f, identifierTranslations[i], bits=32)

    align(f)

    # Segment 4: StringTable
    stringTableEntries = obj["stringTableEntries"]
    for i in range(header["stringCount"]):
        for key in stringTableEntryS:
            stringTableEntry = stringTableEntries[i]
            write(f, stringTableEntry[key], stringTableEntryS[key])

    align(f)

    # Segment 5: StringTableOverflow
    stringTableOverflowEntries = obj["stringTableOverflowEntries"]
    for i in range(header["overflowStringCount"]):
        for key in overflowStringTableEntryS:
            stringTableOverflowEntry = stringTableOverflowEntries[i]
            write(f, stringTableOverflowEntry[key], overflowStringTableEntryS[key])

    align(f)

    # Segment 6: StringStorage
    stringStorage = obj["stringStorage"]
    stringStorageS[2] = header["stringStorageSize"]
    write(f, stringStorage, stringStorageS)

    align(f)

    # Segment 7: ArrayBuffer
    arrayBuffer = obj["arrayBuffer"]
    arrayBufferS[2] = header["arrayBufferSize"]
    write(f, arrayBuffer, arrayBufferS)

    align(f)

    # Segment 9: ObjKeyBuffer
    objKeyBuffer = obj["objKeyBuffer"]
    objKeyBufferS[2] = header["objKeyBufferSize"]
    write(f, objKeyBuffer, objKeyBufferS)

    align(f)

    # Segment 10: ObjValueBuffer
    objValueBuffer = obj["objValueBuffer"]
    objValueBufferS[2] = header["objValueBufferSize"]
    write(f, objValueBuffer, objValueBufferS)

    align(f)

    # Segment 11: RegExpTable
    regExpTable = obj["regExpTable"]
    for i in range(header["regExpCount"]):
        regExpEntry = regExpTable[i]
        for key in regExpTableEntryS:
            write(f, regExpEntry[key], regExpTableEntryS[key])

    align(f)    
    
    # Segment 12: RegExpStorage
    regExpStorage = obj["regExpStorage"]
    regExpStorageS[2] = header["regExpStorageSize"]
    write(f, regExpStorage, regExpStorageS)

    align(f)

    # Segment 13: CJSModuleTable
    cjsModuleTable = obj["cjsModuleTable"]
    for i in range(header["cjsModuleCount"]):
        cjsModuleEntry = cjsModuleTable[i]
        for key in cjsModuleTableS:
            write(f, cjsModuleEntry[key], cjsModuleTableS[key])
        
    align(f)

    # Write remaining
    f.writeall(obj["inst"])

    # Write Overflowed Function Header
    for overflowedFunctionHeader in overflowedFunctionHeaders:
        smallFunctionHeader = overflowedFunctionHeader["small"]
        large_offset = (smallFunctionHeader["infoOffset"] << 16 )  | smallFunctionHeader["offset"]
        f.seek(large_offset)
        for key in functionHeaderS:
            write(f, overflowedFunctionHeader[key], functionHeaderS[key])

